/*
 * Copyright (c) 2004 Nokia. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of Nokia nor the names of its contributors may be
 * used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "Cookie.h"
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

Cookie::Cookie(const gchar* name, const gchar* value, bool)
	:m_name(name)
	,m_value(value)
	,m_version(-1)
	,m_maxAge(-1)
	,m_secure(false)
	,m_discard(false)
	,m_ports(0)
	,m_nports(-1)
	,m_forceInvalid(false)
{
}

Cookie::~Cookie()
{
	if (m_ports)
		delete[] m_ports;
}

bool Cookie::isValidForURL(GURI* gurl)
{
	if (m_forceInvalid)
		return false;

   /* Reject if: (from rfc2965) */

   /*
	*  The value for the Path attribute is not a prefix of the
	*  request-URI.
	*
	*  FIXME: $Path=/acme/login  matches for url www.example.com/acme/loginner/ ??
	*/
	if (m_path)
		if (!g_str_has_prefix(gnet_uri_get_path(gurl), m_path))
			return false;

	if (m_domain) {
		/*
		 *  The value for the Domain attribute contains no embedded dots,
		 * and the value is not .local.
		 */
		/*
		 *  The effective host name that derives from the request-host does
		 *  not domain-match the Domain attribute.
		 */
		if (strcmp(".local", m_domain) && !strchr(m_domain.ptr() + 1, '.'))
			return false;

		/*
		 * The request-host is a HDN (not IP address) and has the form HD,
         * where D is the value of the Domain attribute, and H is a string
         * that contains one or more dots.
		 */
		// fixme: is not ip addr
		gchar* hostname = gurl->hostname;

		int hostnameLen = strlen(hostname);
		int domainLen = strlen(m_domain);
		int prefixLen = hostnameLen - domainLen;

		// check for suffix. Since we calculate lens anyway, we can do this by hand
		// instead of g_has_suffix
		if (hostnameLen < domainLen || strcmp(hostname+prefixLen, m_domain))
			return false; 	// not explicitly stated

		for (int i = 0; i < prefixLen; ++i)
			if (hostname[prefixLen] == '.')
				return false;
	}

	/*
	 * The Port attribute has a "port-list", and the request-port was
	 * not in the list.
	 */
	if (m_ports) {
		int port = gurl->port;
		if (!port)
			port = 80;

		for (int i = 0; i  < m_nports; ++i) {
			if (m_ports[i] == port)
				goto portFound;		// this is legimite use of goto. really it is.
		}
		return false;
	}
	portFound:

    return true;
}

bool Cookie::isExpired()
{
	if (m_maxAge == 0)
		return true;
    return false;
}


void Cookie::setInvalid()
{
	m_forceInvalid = true;
}


const gchar* Cookie::headerString()
{
	if (m_headerString)
		return m_headerString;

	// generate outgoing Cookie: -header from this cookie.
	GString* buffer = g_string_sized_new (32);
	bool quote = m_version == 1;

	g_string_append_printf(buffer,
						   quote ? "%s=\"%s\"": "%s=%s", m_name.ptr(), m_value.ptr());
	if (m_path)
		g_string_append_printf(buffer, ";$Path=\"%s\"", m_path.ptr());

	if (m_domain)
		g_string_append_printf(buffer, quote?";$Domain=\"%s\"":";$Domain=%s", m_domain.ptr());
	m_headerString.adopt(buffer->str);
	g_string_free(buffer, FALSE);
	return m_headerString;
}
