/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* Copyright (C) 2007 Kimmo Kinnunen<kimmok@iki.fi>
 * bsd license
 */
#include "CookieParser.h"

#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include "Cookie.h"
CookieParser::CookieParser()
	:m_ncookies(0)
	,m_nports(0)
	,m_err(false)
{

}
CookieParser::~CookieParser()
{
    deleteCookies();
}

void CookieParser::deleteCookies()
{
	for(int i = 0; i < m_ncookies; ++i)
		delete (m_cookies[i]);
	m_ncookies = 0;
	m_curCookie = 0;
}

#define SKIPWS	for (;*m_value && (*m_value == ' '|| *m_value == '\t' || *m_value == '\r'|| *m_value == '\n'); ++m_value) /*empty*/;

void CookieParser::readNext(gchar ch)
{
	if (m_err) return;
	SKIPWS;

	if (!ifNext(ch))
		m_err = true;
}



bool CookieParser::ifNext(gchar ch)
{
	if (m_err) return false;

	gchar* v = m_value;
	SKIPWS;

	if (*m_value == ch) {
		++m_value;
		return true;
	}
	m_value = v;
	return false;
}

/*
#!/usr/bin/perl
my $notok = "()<>@,;:\\'/[]?={} \t\"";
my $n = 0;
for (my $i=0; $i < 127; ++$i) {
	my $letter = chr($i);
	$n |= (1 << ($i % 8)) if (index($notok, $letter) == -1 && ($i > 31) && $i < 127);
    if (($i % 8) == 7) {
	    printf "0x%02x,", $n;
		$n = 0;
    }
}
printf "0x%02x", $n if (($i % 8) != 7);
print "\n";
*/
static char tokenChars[] = {0x00,0x00,0x00,0x00,0x7a,0x6c,0xff,0x03,0xfe,0xff,0xff,0xc7,0xff,0xff,0xff,0x57};

inline bool isValidTokenChar(gchar c) {
	return (c<127) && (tokenChars[c>>3] & (1u << (c % 8u)));
}

gchar* CookieParser::readToken(bool isPath)
{
	if (m_err) return 0;
	SKIPWS;
	gchar* start = m_value;
	for(; isValidTokenChar(*m_value) || (isPath && *m_value=='/'); ++m_value) /*empty*/;
	if (start != m_value)
		return g_strndup(start, m_value - start);

	m_err = true;
	return 0;
}

gchar* CookieParser::readValue(bool isPath)
{
	if (m_err) return 0;
	SKIPWS;

	if (ifNext('"')) {
		return readQuote();
	}

	return readToken(isPath);
}

gchar* CookieParser::readQuote()
{
	if (m_err) return 0;
	SKIPWS;

	gchar* start = m_value;
	gchar* end = strchr(m_value, '"');
	if (end) {
		m_value = end+1;
		return g_strndup(start, end-start);
	}
	m_err = true;
	return 0;
}

void CookieParser::readUnknown()
{
	if (m_err) return;
	SKIPWS;
	for(; *m_value && *m_value != ';';++m_value) /*empty*/;

	return;
}

int CookieParser::readPlainInt()
{
	if (m_err) return -1;
	errno = 0;
	gchar* endptr = 0;
	long res = strtol(m_value, &endptr, 10);
	if (!errno && endptr && endptr != m_value) {
		m_value = endptr;
		return res;
	}

	m_err = true;
	return -1;
}

int CookieParser::readInt()
{
	if (m_err) return -1;

	gchar ch = ';';
	if (*m_value == '"') {
		ch = '"';
		++m_value;
	}
	
	int res = readPlainInt();

	if (!m_err) {
		if (ch=='"') {
			if (*m_value != '"') {
				m_err = true;
				return -1;
			}
			++m_value;
		}
		return (int)res;
	}

	m_err = true;
	return -1;
}

#define TOKENEQ(str) ((g_ascii_strcasecmp((str), token)) == 0)

CookieParser::State CookieParser::readStateToken()
{
	if (m_err) return ERROR;

	SKIPWS;
	own_gcharptr token;
	token.adopt(readToken());
	if (!token)
		return ERROR;

	if (TOKENEQ("version")) {
		return VERSION;
	}else if (TOKENEQ("comment")) {
		return COMMENT;
	} else if (TOKENEQ("commenturl")) {
		return COMMENT_URL;
	} else if (TOKENEQ("discard")) {
		return DISCARD;
	} else if (TOKENEQ("domain")) {
		return DOMAIN;
	} else if (TOKENEQ("max-age")) {
		return MAXAGE;
	} else if (TOKENEQ("path")) {
		return PATH;
	} else if (TOKENEQ("port")) {
		return PORT;
	} else if (TOKENEQ("secure")) {
		return SECURE;
	}

	return UNKNOWN;
}

void CookieParser::emitNameValue(gchar* name, gchar* value)
{
	m_curCookie = m_cookies[m_ncookies++] = new Cookie(name, value, true);

}

void CookieParser::emitPortListStart()
{
	m_nports = 0;
}

void CookieParser::emitPortListEnd(bool error)
{
	if (!error && !m_curCookie->m_ports) {
		m_curCookie->m_ports = new int[m_nports];
		m_curCookie->m_nports = m_nports;
		for (int i=0;i < m_nports; ++i)
			m_curCookie->m_ports[i] = m_ports[i];
	}
	m_nports = 0;
}

void CookieParser::emitPort(int port)
{
	if (m_nports < s_portsSz)
		m_ports[m_nports++] = port;
}

void CookieParser::emitVersion(int version)
{
	if (m_curCookie->m_version == -1)
		m_curCookie->m_version = version;
}

void CookieParser::emitMaxAge(int age)
{
	if (m_curCookie->m_maxAge == -1)
		m_curCookie->m_maxAge = age;
}

void CookieParser::emitComment(gchar* comment)
{
	if (!m_curCookie->m_comment)
		m_curCookie->m_comment = comment;
}

void CookieParser::emitCommentURL(gchar* url)
{
	if (!m_curCookie->m_commentURL)
		m_curCookie->m_commentURL = url;
}

void CookieParser::emitPath(gchar* path)
{
	if (!m_curCookie->m_path)
		m_curCookie->m_path = path;
}

void CookieParser::emitDomain(gchar* domain)
{
	if (!m_curCookie->m_domain)
		m_curCookie->m_domain = domain;
}

void CookieParser::emitSecure()
{
	m_curCookie->m_secure = true;
}

void CookieParser::emitDiscard()
{
	m_curCookie->m_discard = true;
}

void CookieParser::emitUnknown()
{
}

Cookie** CookieParser::parse(const gchar* cookies)
{
	deleteCookies();

	m_value = const_cast<gchar*>(cookies);

	State state = COOKIE;

	while(state != END && !m_err) {
		if (m_ncookies == s_cookiesSz-1)
			break;

		switch(state) {
		case MAYBE_COOKIE: {
			if (ifNext(',')) {
				state = COOKIE;
			} else
                state = END;
			break;
		}

		case COOKIE: {
			gchar* n = readToken();
			readNext('=');
			gchar* v = readValue();

			emitNameValue(n,v);
			state = MAYBE_COOKIE_AV;
			break;
		}

		case MAYBE_COOKIE_AV: {
			if (ifNext(';')) {
				state = COOKIE_AV;
			} else
                state = MAYBE_COOKIE;
			break;
		}
		case COOKIE_AV: {
			State stateToken = readStateToken();
			state = stateToken;
			break;
		}
		case COMMENT: {
			readNext('=');
			gchar* t = readValue();
			emitComment(t);
            state = MAYBE_COOKIE_AV;
			break;
		}
		case COMMENT_URL: {
			readNext('=');
			gchar* t = readValue();
			emitCommentURL(t);
            state = MAYBE_COOKIE_AV;
			break;
		}
		case DOMAIN: {
			readNext('=');
			gchar* t = readValue();
			emitDomain(t);
            state = MAYBE_COOKIE_AV;
			break;
		}
		case MAXAGE: {
			readNext('=');

			int age = readInt();
			emitMaxAge(age);
            state = MAYBE_COOKIE_AV;
			break;
		}
		case PATH: {
			readNext('=');
			gchar* t = readValue(true);
			emitPath(t);
            state = MAYBE_COOKIE_AV;
			break;
		}
		case PORT: {
			readNext('=');
			emitPortListStart();
            state = PORT_LIST_QUOTE;
			break;
		}
		case SECURE: {
			emitSecure();
			state = MAYBE_COOKIE_AV;
			break;
		}
		case DISCARD: {
			emitDiscard();
			state = MAYBE_COOKIE_AV;
			break;
		}
		case VERSION: {
			readNext('=');
			int i = readInt();
			emitVersion(i);
			state = MAYBE_COOKIE_AV;
			break;
		}
		case UNKNOWN: {
			readUnknown();
			emitUnknown();
			state = MAYBE_COOKIE_AV;
			break;
		}

		case PORT_LIST_QUOTE: {
			readNext('"');
			state = PORT_LIST;
			break;
		}
		case PORT_LIST: {
			do {
				int i = readPlainInt();
				emitPort(i);
			} while (ifNext(','));

			emitPortListEnd(!ifNext('"'));
			state = MAYBE_COOKIE_AV;
			break;
		}
		default: {
            m_err = true;
			break;
		}
		}
	}

	m_value = 0;
	m_cookies[m_ncookies] = 0;
	return m_cookies;
}
