/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * Copyright (c) 2004 Nokia. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce theb above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of Nokia nor the names of its contributors may be
 * used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <assert.h>
#include <string.h> // strcmp

#include "Cookie.h"
#include "CookieStorage.h"
#include "CookieParser.h"
#include "uri.h"

WebCoreCookieAdapter* WebCoreCookieAdapter::sharedAdapter()
{
    return CookieStorage::shared();
}

CookieStorage* CookieStorage::shared()
{
    static CookieStorage single;
    return &single;
}


/** Class storing all cookies from one cookie domain
 * cookie domain is not a real domain.
 */
class CookieDomain
{
public:
    CookieDomain(Cookie*, GURI*);
    ~CookieDomain();

    const gchar* getCookieHeaderLines(GURI* uri);

    /** ownership is transferred */
    void setCookie(Cookie* cookie);

    void removeAllExpired();

    /** ownership not transferred*/
    gchar* key() { return const_cast<gchar*>(m_domainkey.ptr()); }


    static gchar* createKey(Cookie*, GURI*);
private:
    own_gcharptr m_domainkey;
    own_gcharptr m_cookieBuffer;
    GSList* m_cookies;
};

extern "C" {
static void deleteCookie(gpointer data, gpointer /*user_data*/)
{
    assert(data);
    Cookie* c = static_cast<Cookie*>(data);
    delete(c);
}

// ugly
struct ConcatParams {
    GString* buffer;
    GURI* uri;
};

static void concatenateCookiesV1(gpointer data, gpointer user_data)
{
    Cookie* cookie = static_cast<Cookie*>(data);
    ConcatParams* cp = static_cast<ConcatParams*>(user_data);
    if (cookie->version() == 1  && cookie->isValidForURL(cp->uri)) {
        if (cp->buffer->len == 0) {
            g_string_append(cp->buffer, "Cookie: $Version=\"1\";");
        }
        g_string_append(cp->buffer, cookie->headerString());
        g_string_append(cp->buffer, ";");
    }
}

static void concatenateCookiesV0(gpointer data, gpointer user_data)
{
    Cookie* cookie = static_cast<Cookie*>(data);
    ConcatParams* cp = static_cast<ConcatParams*>(user_data);
    if ((cookie->version() == 0 || cookie->version() == -1) && cookie->isValidForURL(cp->uri)) {
        if (cp->buffer->len == 0) {
            g_string_append(cp->buffer, "Cookie: ");
        }
        g_string_append(cp->buffer, cookie->headerString());
        g_string_append(cp->buffer, ";");
    }
}

static gint compareCookie(gconstpointer a, gconstpointer b)
{
    // cookies match if
    // - cookie domains (checked previously via hash table)
    // - domain and path match
    // - name matches

    const Cookie* ca = static_cast<const Cookie*>(a);
    const Cookie* cb = static_cast<const Cookie*>(b);
    // compare domains case-insensitively, watch out for nulls
    if (!ca->domain() || !cb->domain()) {
        if (ca->domain()) return 1;
        if (cb->domain()) return -1;
    } else {
        int r = g_ascii_strcasecmp(ca->domain(), cb->domain());
        if (r)
            return r;
    }
    // compare paths case-sensitively, watch out for nulls
    if (!ca->path() || !cb->path()) {
        if (ca->path()) return 1;
        if (cb->path()) return -1;
    } else {
        int r = strcmp(ca->path(), cb->path());
        if (r)
            return r;
    }

    assert (ca->name());
    assert (cb->name());

    // match if returns 0
    // Fixme: verify that this is case-sensitive
    return strcmp(ca->name(), cb->name());
}
}

CookieDomain::CookieDomain(Cookie* cookie, GURI* guri)
    :m_cookies(NULL)
{
    gchar* key = createKey(cookie, guri);
    m_domainkey.adopt(key);

    setCookie(cookie);
}

CookieDomain::~CookieDomain()
{
    g_slist_foreach(m_cookies, deleteCookie, NULL);
    g_slist_free(m_cookies);
}

gchar* CookieDomain::createKey(Cookie* cookie, GURI* guri)
{
    const gchar* d = cookie->domain();
    if (!d)
        d = guri->hostname;

    gchar* k = g_strdup(d);

    return k;
}


void CookieDomain::setCookie(Cookie* cookie)
{
    m_cookieBuffer = 0;
    GSList* entry = g_slist_find_custom(m_cookies, cookie, compareCookie);
    if (entry) {
        const Cookie* ca = static_cast<const Cookie*>(entry->data);
        delete (ca);
        entry->data = cookie;
    } else {
        m_cookies = g_slist_append(m_cookies, cookie);
    }
}

void CookieDomain::removeAllExpired()
{
    bool someRemoved = false;
    GSList* iter = m_cookies;
    while (iter) {
        Cookie* cookie = static_cast<Cookie*>(iter->data);
        if (cookie->isExpired()) {
            someRemoved = true;
            delete (cookie);
            iter->data = NULL;  // mark as removed
        }
        iter = iter->next;
    }

    if (someRemoved) {
        m_cookieBuffer = 0;
        m_cookies = g_slist_remove_all(m_cookies, NULL);
    }
}

const gchar* CookieDomain::getCookieHeaderLines(GURI* uri)
{
    removeAllExpired();

    if (!m_cookies)
		return 0L;
#if 0
    if (m_cookieBuffer)
        return m_cookieBuffer;
#endif
    ConcatParams cp;
    cp.uri = uri;

    GString* bufferv1 = g_string_sized_new (32);
    cp.buffer = bufferv1;
    g_slist_foreach(m_cookies, concatenateCookiesV1, &cp);

    GString* bufferv0 = g_string_sized_new (32);
    cp.buffer = bufferv0;
    g_slist_foreach(m_cookies, concatenateCookiesV0, &cp);

    if (bufferv1->len) {
        gchar* sc = strrchr(bufferv1->str, ';');
        if (sc) *sc = '\0';
        bufferv1->len = sc - bufferv1->str;
        if (bufferv0->len) {
            g_string_append(bufferv1, "\r\n");
            g_string_append(bufferv1, bufferv0->str);
            sc = strrchr(bufferv1->str, ';');
            if (sc) *sc = '\0';
            bufferv1->len = sc - bufferv1->str;
        }
        m_cookieBuffer.adopt(bufferv1->str);
        g_string_free (bufferv1, FALSE);
        g_string_free (bufferv0, TRUE);

    } else if (bufferv0->len) {
        gchar* sc = strrchr(bufferv0->str, ';');
        if (sc) *sc = '\0';

        m_cookieBuffer.adopt(bufferv0->str);
        g_string_free(bufferv1, TRUE);
        g_string_free(bufferv0, FALSE);

    } else {
        m_cookieBuffer = 0;
    }

    return m_cookieBuffer;
}

extern "C" {
void deleteCookieDomain(gpointer data)
{
    CookieDomain* cj = static_cast<CookieDomain*>(data);
    delete (cj);
}
}

/**
 * Dummy implementation for cookie storage
 * Stores cookie according to url,
 * Is not persistent
 */

CookieStorage::CookieStorage()
    :m_enabled(true)
    ,m_cookieBuffer(0)
{
    m_cookiesByURL = g_hash_table_new_full(g_str_hash,
                                           g_str_equal,
                                           NULL,
                                           deleteCookieDomain);
}

CookieStorage::~CookieStorage()
{
    g_hash_table_destroy(m_cookiesByURL);
}

bool CookieStorage::cookiesEnabled()
{
    return m_enabled;
}

void CookieStorage::setCookiesEnabled(bool flag)
{
	m_enabled = flag;
}


// for each parent subdomain (ascending the domain hierarchy)


const gchar* CookieStorage::cookiesForURL(const gchar* URL)
{
	assert(URL);

    if (!m_enabled) return 0L;

    m_cookieString = 0;

    GURI* uri = gnet_uri_new(URL);

    if (!uri->hostname)
		return 0L;

    const int slashesSz = 256;
    int nslashes=0;
    gchar* slashes[slashesSz];

    const int dotsSz = 256;
    int ndots = 0;
    gchar* dots[dotsSz];

    gchar* key = g_strconcat(uri->hostname, gnet_uri_get_path(uri), NULL);
    gchar* firstSlash = key + strlen(uri->hostname);

    dots[ndots++] = key;        // first the most specific, the full doman name

    for (gchar* i = key; i < firstSlash; ++i) {
        if (*i == '.') {
            dots[ndots++] = i;
        }

        if (ndots == dotsSz)
            break;
    }

    for (gchar* i = firstSlash ; *i != '\0'; ++i) {
        if (*i == '/')
            slashes[nslashes++] = i;

        if (nslashes == slashesSz)
            break;
    }

    // watch out that we didn't overflow
    if (nslashes == slashesSz || ndots == dotsSz) {
        g_free(key);
        gnet_uri_delete(uri);
        return 0;
    }

    *(slashes[0]) = '\0';
    //char tmp;
    // for each parent subdomain
    for (int d = 0; d < ndots; ++d) {
        gchar* curkey = dots[d];
        appendCookieStringForKey(curkey, uri);
    }

    g_free(key);
    gnet_uri_delete(uri);

    if (m_cookieBuffer) {
        m_cookieString.adopt(m_cookieBuffer->str);
        g_string_free(m_cookieBuffer, FALSE);
        m_cookieBuffer = 0;
    }

    return m_cookieString;
}

void CookieStorage::appendCookieStringForKey(const gchar* cookieDomain, GURI* uri)
{
    CookieDomain* cj = static_cast<CookieDomain*>(g_hash_table_lookup(m_cookiesByURL, cookieDomain));

    if (!cj)  return;

    const gchar* str = cj->getCookieHeaderLines(uri);

    if (str) {
        if (!m_cookieBuffer)
            m_cookieBuffer = g_string_sized_new (32);

        g_string_append(m_cookieBuffer, str);
    }
}

void CookieStorage::setValidCookieForURL(Cookie* cookie, GURI* uri)
{
    gchar* key = CookieDomain::createKey(cookie, uri);

    CookieDomain* cj = static_cast<CookieDomain*>(g_hash_table_lookup(m_cookiesByURL, key));
    if (cj) {
        cj->setCookie(cookie);
    } else {
        cj = new CookieDomain(cookie, uri);
        g_hash_table_insert(m_cookiesByURL, cj->key(), cj);
    }
    g_free(key);
}

void CookieStorage::setCookiesForURL(const gchar* cookies,
									 const gchar* URL,
									 const gchar* policyBaseURL)
{
	assert(cookies);
	assert(URL);
	//assert(policyBaseURL);

    if (!m_enabled) return;

    CookieParser p;
    Cookie** cv = p.parse(cookies);
    GURI* uri = gnet_uri_new(URL);

    for (int i = 0; cv[i]; ++i) {
        Cookie* cookie = cv[i];
        if (cookie->isValidForURL(uri)) {
            setValidCookieForURL(cookie, uri);
            cv[i] = 0;
        }
    }

    gnet_uri_delete(uri);
}

