/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * Copyright (c) 2004 Nokia. All rights reserved.
 * Copyright (c) 2007 Kimmo Kinnunen <kimmok@iki.fi>. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of Nokia nor the names of its contributors may be
 * used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "PageLoadListener.h"
#include <string.h>

#include "Cookie.h"
#include "ResourceLoadListener.h"
#include "loader/MIMETypeResolver.h"
#include "gtk/webi-resource-private.h"

static bool canSupportNatively(const gchar* contentType)
{
	if ((strcmp(contentType, "text/html") == 0)
		|| (strcmp(contentType, "text/plain") == 0)
		|| (strcmp(contentType, "text/xml") == 0)
		|| (strcmp(contentType, "application/xhtml+xml") == 0))
		return true;
	else if ((strcmp(contentType, "image/jpeg") == 0)
			 || (strcmp(contentType, "image/gif") == 0)
			 || (strcmp(contentType, "image/png") == 0))
		return true;

	return false;
}

/** The default implementation of functionality that feeds page load data data to the engine.
 */
class DataAdder
{
public:
	DataAdder(BridgeImpl* bridge) : m_bridge(bridge) {}
	virtual ~DataAdder() {}

	virtual void openURL(const gchar* URL, bool reload, const gchar* contentType, const gchar* refresh, GTime lastModified, KWIQPageCache* pageCache);
	virtual void setEncoding(const gchar* encoding, bool gotEncoding);
	virtual void addData(const char* data, int len);
	virtual void loadURLFinished(bool gotError);

protected:
	BridgeImpl* m_bridge;
};

void DataAdder::openURL(const gchar* URL, bool reload, const gchar* contentType, const gchar* refresh, GTime lastModified, KWIQPageCache* pageCache)
{
    m_bridge->openURL(URL, reload, contentType, refresh, lastModified, pageCache);
}

void DataAdder::setEncoding(const gchar* encoding, bool gotEncoding)
{
    m_bridge->setEncoding(encoding, !gotEncoding);
}

void DataAdder::addData(const char* data, int len)
{
	m_bridge->addData(data, len);
}

void DataAdder::loadURLFinished(bool gotError)
{
	m_bridge->loadURLFinished(gotError);
}

/** Transforms data loaded with mime-type text/plain to be html which contains transferred text.
 *  The text is formatted using <pre> -tag
 */
class TextPlainDataAdder : public DataAdder
{
public:
	TextPlainDataAdder(BridgeImpl* bridge, const gchar* URL)
		:DataAdder(bridge)
		,m_prologueSent(false)
		,m_url(URL)
		{}

	~TextPlainDataAdder() {}

	void openURL(const gchar* URL, bool reload, const gchar* contentType, const gchar* refresh, GTime lastModified, KWIQPageCache* pageCache);
	void setEncoding(const gchar* encoding, bool gotEncoding);
	void addData(const char* data, int len);
	void loadURLFinished(bool gotError);

	static const gchar* prologue1;
	static const gchar* prologue2;
	static const gchar* epilogue1;
public:
	bool m_prologueSent;
	own_gcharptr m_url;
};

const gchar* TextPlainDataAdder::prologue1 = "<html><head><title>";
const gchar* TextPlainDataAdder::prologue2 = "</title><body><pre>";
const gchar* TextPlainDataAdder::epilogue1 = "</pre></body></html>";


void TextPlainDataAdder::openURL(const gchar* URL, bool reload, const gchar* contentType, const gchar* refresh, GTime lastModified, KWIQPageCache* pageCache)
{
    m_bridge->openURL(URL, reload, "text/html", refresh, lastModified, pageCache);
}

void TextPlainDataAdder::setEncoding(const gchar* encoding, bool gotEncoding)
{
    m_bridge->setEncoding(encoding, !gotEncoding);
}


void TextPlainDataAdder::addData(const char* data, int len)
{
	int dataLen = 0;
    if (!m_prologueSent) {
        gchar* c = g_strconcat(prologue1, m_url.ptr(), prologue2, NULL);
		dataLen = strlen(c);
		m_bridge->addData(c, dataLen);
		g_free(c);
        m_prologueSent = true;
    }

    m_bridge->addData(data, len);
}

void TextPlainDataAdder::loadURLFinished(bool gotError)
{
	m_bridge->addData(epilogue1, sizeof(epilogue1) - 1); // no need to send null byte
	m_bridge->loadURLFinished(gotError);
}

/** Transforms plain image loads to html page, which reference the same image url
 * ends up loading image data twice. Should be changed somehow
 */
class ImageDataAdder : public DataAdder
{
public:
	ImageDataAdder(BridgeImpl* bridge, const char* URL)
		:DataAdder(bridge)
		,m_prologueSent(false)
		,m_url(URL)
		{}

	~ImageDataAdder() {}

	void openURL(const gchar* URL, bool reload, const gchar* contentType, const gchar* refresh, GTime lastModified, KWIQPageCache* pageCache);
	void setEncoding(const gchar* encoding, bool gotEncoding);
	void addData(const char* data, int len);
	void loadURLFinished(bool gotError);

	static const gchar* prologue1;
	static const gchar* prologue2;
	static const gchar* prologue3;

private:
	bool m_prologueSent;
	own_gcharptr m_url;
};

const gchar* ImageDataAdder::prologue1 = "<html><head><title>Image: ";
const gchar* ImageDataAdder::prologue2 = "</title><body><img src=\"";
const gchar* ImageDataAdder::prologue3 = "\"></img></body></html>";

void ImageDataAdder::openURL(const gchar* URL, bool reload, const gchar* contentType, const gchar* refresh, GTime lastModified, KWIQPageCache* pageCache)
{
    m_bridge->openURL(URL, reload, "text/html", refresh, lastModified, pageCache);
}

void ImageDataAdder::setEncoding(const gchar* encoding, bool gotEncoding)
{
    m_bridge->setEncoding("utf-8", false);
}

void ImageDataAdder::addData(const char* data, int len)
{
	int dataLen = 0;
    if (!m_prologueSent) {
        gchar* c = g_strconcat(prologue1, m_url.ptr(), prologue2, m_url.ptr(), prologue3, NULL);
		dataLen = strlen(c);
		m_bridge->addData(c, dataLen);
		g_free(c);
        m_prologueSent = true;
    }
}

void ImageDataAdder::loadURLFinished(bool gotError)
{
	m_bridge->loadURLFinished(gotError);
}

static DataAdder* newDataAdderForContentType(const gchar* contentType, BridgeImpl* bridge, const gchar* URL)
{
	if (!strcmp(contentType, "text/plain"))
		return new TextPlainDataAdder(bridge, URL);
	else if (g_str_has_prefix(contentType, "image/") == TRUE)
		return new ImageDataAdder(bridge, URL);
	return new DataAdder(bridge);
}

/** Handles the html page loads.
 * "Page load" is the operation when url is loaded to the window. The page load resource
 * is usually html file fetched from network. This html file references other resources,
 * but they're loaded in ResourceLoader.
 */
PageLoadListener::PageLoadListener(BridgeImpl* b, const gchar * URL)
    :m_bridge(b)
	,m_url(URL)
	,m_contentLength(0)
    ,m_URLOpen(false)
    ,m_gotEncoding(false)
    ,m_gotError(false)
    ,m_committed(false)
	,m_contentHandlerKnown(false)
	,m_request(0)
    ,m_dataAdder(0)
{
    m_lastModified = 0;
}

PageLoadListener::~PageLoadListener()
{
	delete(m_request);
}

void PageLoadListener::externalLoadCancelled()
{
	delete(this);
}

void PageLoadListener::commit()
{
	// load should be committed when first data arrives
	// this can be headers or just the data
	if (!m_committed) {
		m_bridge->commitLoad();
		m_committed = true;
	}
}

bool PageLoadListener::isNativeContent()
{
	if (!contentHandlerKnown())
		determineContentHandler();

	return m_bridge;
}

void PageLoadListener::determineContentType() {
	if (!m_contentType)
		m_contentType = Loader::MIMETypeResolver::sharedFactory().mimeTypeForURL(m_url);
}

void PageLoadListener::determineContentHandler()
{

    determineContentType();

	if (!canSupportNatively(m_contentType)) {
		m_bridge->handleLoadExternally(this);
		m_externalDelegate = m_bridge->getDefaultExternalContentDelegate();

		// This marks the page load as "non-native"
		m_bridge = 0;

		if (m_externalDelegate) {
			m_externalResource.adopt(webi_resource_new(this, m_url, m_contentType));
			if (m_contentLength)
				m_externalResource->content_length = m_contentLength;

			webi_external_content_delegate_start(m_externalDelegate, m_externalResource);
		}
	}

	m_contentHandlerKnown = true;
}

void PageLoadListener::data(const Loader::Request*, const char* data, int len)
{
	if (isNativeContent()) {
		if (!isURLOpen())
			openURL();

		m_dataAdder->addData(data, len);

		m_status.addReceived(len);
		m_bridge->emitResourceLoadStatus(&m_status);
	} else {
		if (m_externalDelegate)
			webi_external_content_delegate_data(m_externalDelegate, m_externalResource, data, len);
	}
}

void PageLoadListener::started(const Loader::Request*)
{
    m_bridge->emitResourceLoadStarted(&m_status);
}

void PageLoadListener::finished(const Loader::Request*)
{
	if ((m_gotError && !contentHandlerKnown()) || isNativeContent()) {
		if (!isURLOpen())
			openURL();

		m_dataAdder->loadURLFinished(m_gotError);
		m_bridge->emitResourceLoadFinished(&m_status);

		if ( m_bridge->numPendingOrLoadingRequests() == 0 || m_gotError)
			m_bridge->emitFrameLoadFinished(m_gotError?-1:0);

	} else {
		if (m_externalDelegate)
			webi_external_content_delegate_end(m_externalDelegate, m_externalResource, m_gotError);

		m_externalResource = 0;
	}
}

void PageLoadListener::error(const Loader::Request*)
{
    m_status.setError();
    m_gotError = true;
}

void PageLoadListener::openURL()
{
	if (!isCommitted())
		commit();

    // on error we need m_contentType. 
    determineContentType();

    m_bridge->emitResourceLoadHeaders(&m_status);
    m_bridge->saveHistoryItem();

	m_dataAdder = newDataAdderForContentType(m_contentType, m_bridge, m_url);

    /* clear current state */
	m_dataAdder->openURL(m_url, false, m_contentType, m_refresh, m_lastModified, 0);
	m_dataAdder->setEncoding(m_encoding, !m_gotEncoding);

    m_URLOpen = true;
}

void PageLoadListener::setHTTPStatus(const Loader::Request*, int status)
{
	if (!isCommitted())
		commit();

}

void PageLoadListener::setLocation(const Loader::Request*, const gchar* absoluteURL)
{
    m_bridge->redirection(absoluteURL);
	m_url = absoluteURL;
}

void PageLoadListener::setContentType(const Loader::Request*, const gchar* contentType, const gchar* encoding)
{
    if (encoding) {
        m_encoding = encoding;
        m_gotEncoding = true;
    } else {
        m_encoding = 0;
    }

    if (contentType)
        m_contentType = contentType;
}

void PageLoadListener::setContentLength(const Loader::Request*, unsigned long contentLength)
{
	m_contentLength = contentLength;
    m_status.setSize(contentLength);
}

void PageLoadListener::setCookie(const Loader::Request*, const gchar* value)
{
    m_bridge->emitSetCookie(value, m_url);
}

void PageLoadListener::setRefresh(const Loader::Request*, const gchar* value)
{
    m_refresh = value;
}

void PageLoadListener::setUnknownHTTPHeader(const Loader::Request*, const gchar* key, const gchar* value)
{
}

