/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * Copyright (c) 2007 Hiroyuki Ikezoe All rights reserved.
 * Copyright (c) 2007 Kouhei Sutou All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of authors nor the names of its contributors may be
 * used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <assert.h>
#include <string.h>
#include <pango/pango.h>

#include "PangoTextRendererFactory.h"

WebCoreTextRendererFactory* WebCoreTextRendererFactory::sharedFactory()
{
    static PangoTextRendererFactory singleton;
    return &singleton;
}


PangoNSFont::PangoNSFont(PangoTextRendererFactory* acreator,
             PangoFontDescription* adesc,
             guint ahash)
    : hash(ahash),
      creator(acreator)
{
    assert(adesc);
    desc = adesc;

    spaceWidth = 0;
    xHeight = 0;

    PangoFontMap *map;
    PangoContext *pango_context;
    PangoFont *pango_font;
    PangoRectangle rect;
    PangoAttrList *attrs;
    GList *items, *node;

    pango_context = gdk_pango_context_get();
    map = pango_context_get_font_map(pango_context);
    pango_font = pango_font_map_load_font(map, pango_context, desc);

    const gchar *xText = "x", *spaceText = " ";
    attrs = pango_attr_list_new();
    items = pango_itemize(pango_context, spaceText, 0, strlen(spaceText), attrs, NULL);
    for (node = items; node; node = g_list_next(items)) {
    PangoItem *item = (PangoItem *) items->data;
    PangoGlyphString *glyphs = pango_glyph_string_new();
    pango_shape(&xText[item->offset], item->length, &item->analysis, glyphs);
    pango_glyph_string_extents(glyphs, pango_font, &rect, NULL);
    spaceWidth += rect.width / PANGO_SCALE;
    pango_item_free(item);
    pango_glyph_string_free(glyphs);
    }
    g_list_free(items);

    items = pango_itemize(pango_context, xText, 0, strlen(xText), attrs, NULL);
    for (node = items; node; node = g_list_next(items)) {
    PangoItem *item = (PangoItem *) items->data;
    PangoGlyphString *glyphs = pango_glyph_string_new();
    pango_shape(&xText[item->offset], item->length, &item->analysis, glyphs);
    pango_glyph_string_extents(glyphs, pango_font, &rect, NULL);
    xHeight = rect.height / PANGO_SCALE;
    pango_item_free(item);
    pango_glyph_string_free(glyphs);
    }
    g_list_free(items);

    pango_attr_list_unref(attrs);

    pango_font_get_glyph_extents(pango_font, 0xff, NULL, &rect);
    // FIXEME! set correct information.
    //fixedPitch = pango_font_family_is_monospace(family);
    ascent = PANGO_ASCENT(rect) / PANGO_SCALE;
    descent = PANGO_DESCENT(rect) / PANGO_SCALE;
    lineSpacing = rect.height / PANGO_SCALE;
    g_object_unref(pango_context);
}

PangoNSFont::~PangoNSFont()
{
    pango_font_description_free(desc);

    if (creator)
    creator->fontDeleted(this);
}

extern "C" {
static
void nsfonts_value_detach(gpointer key, gpointer value, gpointer user_data)
{

    PangoNSFont *f = static_cast<PangoNSFont*>(value);
    assert(f);
    f->detach();
}
}

PangoTextRendererFactory::PangoTextRendererFactory()
{
    m_nsfonts = g_hash_table_new(g_direct_hash, g_direct_equal);
}

PangoTextRendererFactory::~PangoTextRendererFactory()
{
    g_hash_table_foreach(m_nsfonts, nsfonts_value_detach, this);
    g_hash_table_destroy(m_nsfonts);
}

NSFont* PangoTextRendererFactory::fontWithFamilies(const gchar * const * families, NSFontTraitMask traits, float size)
{
    assert(families);

    PangoFontDescription *desc;
    desc = pango_font_description_new();

    pango_font_description_set_family(desc, families[0]);
    pango_font_description_set_weight(desc, (traits & NSBoldFontMask) ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
    pango_font_description_set_style(desc, (traits & NSItalicFontMask) ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);

    float dpi = (float) 96.0f;  // this is academic quess
    pango_font_description_set_size(desc, (int)(size * PANGO_SCALE * 72.0f / dpi));

    guint hash = pango_font_description_hash(desc);
    PangoNSFont* font = static_cast<PangoNSFont*>(g_hash_table_lookup(m_nsfonts, GUINT_TO_POINTER(hash)));

    if (!font) {
        font = new PangoNSFont(this, desc, hash);
        g_hash_table_insert(m_nsfonts, GUINT_TO_POINTER(hash), font);
    } else {
        pango_font_description_free(desc);
    }

    return font;
}

PangoTextRenderer* PangoTextRendererFactory::rendererWithFont(NSFont *nsfont, bool isPrinterFont)
{
    assert(nsfont);

    PangoNSFont* font = static_cast<PangoNSFont*>(nsfont);

    return new PangoTextRenderer(this, font);
}

void PangoTextRendererFactory::rendererDeleted(PangoTextRenderer* renderer)
{
}

void PangoTextRendererFactory::fontDeleted(PangoNSFont* font)
{
    g_hash_table_remove(m_nsfonts, GUINT_TO_POINTER(font->hash));
}

bool PangoTextRendererFactory::isFontFixedPitch(NSFont* nsfont)
{
    PangoNSFont* font = static_cast<PangoNSFont*>(nsfont);
    return font->fixedPitch;
}
/*
  vi:ts=4:nowrap:ai:expandtab
*/
