/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * Copyright (c) 2004 Nokia. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of Nokia nor the names of its contributors may be
 * used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "ResourceLoadListener.h"

#include "GLibHelpers.h"

extern "C" {
static void _str_free(gpointer data)
{
	gchar* str = (gchar*) data;
	g_free(str);
}
}

class ResourceLoadResponse : public WebCoreResponse
{
    gchar* _mimeType;
    GHashTable* _headers;
    int _statusCode;

public:
    ResourceLoadResponse()
        :_mimeType(0), _statusCode(-1)  {
        _headers = g_hash_table_new_full(g_str_hash, g_str_equal, _str_free, _str_free);
    }

    ~ResourceLoadResponse() {
        g_hash_table_destroy(_headers);
        if (_mimeType) g_free(_mimeType);
    }

    void setMimeType(const gchar* mime) {
        assignToString(&_mimeType, mime);
    }

    void setStatusCode(int code) {
        _statusCode = code;
    }

    int statusCode() {
        return _statusCode;
    }

    const gchar* mimeType() {
        return _mimeType;
    }

    GHashTable* allHeaderFields() {
        return _headers;
    }

    GTime expiresTime() {
        return 0;
    }

    void addHeader(const gchar* key, const gchar* value) {
        g_hash_table_insert(_headers, g_strdup(key), g_strdup(value));
    }
};


/**
 * Idea of ResourceLoader is to hold WebCoreResourceLoader until corresponding
 * handle is cancelled.
 * after that, WebCoreResourceLoader is released
 */
ResourceLoadListener::ResourceLoadListener(BridgeImpl* b, WebCoreResourceLoader* aloader)
    :m_bridge(b)
    ,m_loader(aloader)
    ,m_response(new ResourceLoadResponse)
    ,m_request(0)
    ,m_status(0)
    ,m_responseSent(false)
    ,m_started(false)
    ,m_cancelled(false)
{
    assert(m_bridge);
    assert(m_loader);
}

ResourceLoadListener::~ResourceLoadListener()
{
}

void ResourceLoadListener::setHTTPStatus(const Loader::Request* request, int astatus)
{
    assert(m_loader);
    assert(!m_responseSent);

    m_response->setStatusCode(astatus);

    m_responseSent = true;
    m_loader->receivedResponse(m_response);
    m_bridge->emitResourceLoadHeaders(&m_status);
}

void ResourceLoadListener::setContentType(const Loader::Request*, const gchar* contentType, const gchar* encoding)
{
    if (contentType)
        m_response->setMimeType(contentType);
}

void ResourceLoadListener::setContentLength(const Loader::Request*, unsigned long contentLength)
{
    m_status.setSize(contentLength);
}

void ResourceLoadListener::setUnknownHTTPHeader(const Loader::Request*, const gchar* key, const gchar* value)
{
    m_response->addHeader(key, value);
}

void ResourceLoadListener::setLocation(const Loader::Request*, const gchar* absoluteURL)
{
}

void ResourceLoadListener::setCookie(const Loader::Request*, const gchar* value)
{
}

void ResourceLoadListener::setRefresh(const Loader::Request*, const gchar* value)
{
}

void ResourceLoadListener::data(const Loader::Request*, const char* data, int len)
{
    assert(m_responseSent);

    m_loader->addData(data, len);
    m_status.addReceived(len);
    m_bridge->emitResourceLoadStatus(&m_status);
}

void ResourceLoadListener::started(const Loader::Request*)
{
    assert(!m_started);
    m_started = true;
    m_bridge->emitResourceLoadStarted(&m_status);
}

void ResourceLoadListener::finished(const Loader::Request* r)
{
    assert(m_loader);

    if (m_response->statusCode() >= 400 || m_status.error())
        m_loader->reportError();
    else
        m_loader->finish();

    m_loader = 0;

    m_bridge->emitResourceLoadFinished(&m_status);
    if (m_bridge->numPendingOrLoadingRequests() == 0) {
        m_bridge->emitFrameLoadFinished(0);
    }
}

void ResourceLoadListener::error(const Loader::Request*)
{
    m_status.setError();
}


WebCoreResourceHandle* ResourceLoadListener::handle(Loader::Request* req)
{
    m_request = req;
    return this;
}

// from WebCoreResourceHandle
void ResourceLoadListener::cancel()
{
    m_loader->cancel();
    m_loader = 0;

    delete(m_request);
    m_request = 0;
}
