/*
 * Copyright (c) 2004 Nokia. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of Nokia nor the names of its contributors may be
 * used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <assert.h>

#include "NRCore/WebCoreCookieAdapter.h"
#include "Cookie.h"
#include "frameimpl.h"
#include "prefimpl.h"
#include "rootimpl.h"
#include "GdkHelpers.h"
#include "GLibHelpers.h"
#include "CookieStorage.h"
#include "CookieParser.h"
#include "webkit-gtk/webkit-frame-private.h"
#include "webkit-gtk/webkit-view-private.h"
#include "webkit-gtk/webkit-url-request.h"
#include "webi-private.h"

class NullDelegate : public OSB::UIDelegate, public OSB::FrameLoadDelegate, public OSB::ResourceLoadDelegate
{
public:

};

static NullDelegate*
nullDelegate()
{
    static NullDelegate single;
    return &single;
}

FrameImpl::FrameImpl(RootImpl* aroot, FrameImpl* parent)
    :BridgeImpl(aroot, parent)
    ,frame_d(nullDelegate())
    ,res_d (nullDelegate())
    ,ui_d(nullDelegate())
    ,_title(0)
    ,_location(0)
    ,_userAgent(0)
{
	m_api = webkit_frame_new(this);
    // must have settings
    initializeSettings(root()->sharedSettings());
}

FrameImpl::~FrameImpl()
{
    if (_title) g_free(_title);
    if (_location) g_free(_location);
    if (_userAgent) g_free(_userAgent);
}


BridgeImpl* FrameImpl::createNewRoot(const gchar* url)
{
	WebkitView* newEngine = 0;

	if (getUIDelegate()){
		WebkitURLRequest* req = webkit_url_request_new_http_get(url);
		newEngine = webkit_ui_delegate_create_view_with_request(getUIDelegate(), root()->api(), req);
		g_object_unref (req);

		if (newEngine) {
			if (WEBKIT_IS_VIEW (newEngine)) {
				return webkit_view_get_root(newEngine)->mainFrame();
			} else if (WEBI_IS_WEBI (newEngine)) {
				// backwards compatibility:
				Webi* webi = WEBI (newEngine);
				return webi->_priv->root->mainFrame();
			}
		}
	}

    return 0;
};

BridgeImpl* FrameImpl::createChildFrame()
{
    FrameImpl* newframe =  new FrameImpl(root(), this);
    newframe->setFrameLoadDelegate(frame_d);
    newframe->setOldUIDelegate(ui_d);
    newframe->setResourceLoadDelegate(res_d);
    return newframe;
}

void FrameImpl::setFrameLoadDelegate(OSB::FrameLoadDelegate* d)
{
    if (d == 0) frame_d = nullDelegate();
    else frame_d = d;
}

void FrameImpl::setResourceLoadDelegate(OSB::ResourceLoadDelegate* d)
{
    if (d == 0) res_d = nullDelegate();
    else res_d = d;
}

void FrameImpl::setOldUIDelegate(OSB::UIDelegate* d)
{
    if (d == 0) ui_d = nullDelegate();
    else ui_d = d;
}


const gchar* FrameImpl::location()
{
    return currentURL();
}

OSB::DeviceType FrameImpl::deviceType() {
    return root()->deviceType();
}

void FrameImpl::startLoad(const char* url)
{
    loadURL(url,
            0,          // referer:
            false,      // reload:
            false,      // onLoadEvent:
            "_self",        // target:
            0,              // triggeringEvent:
            0,          // form:
            0);         // formValues
}

void FrameImpl::refresh()
{
    const gchar* url = requestedURLString();
    loadURL(url,
            0,          // referer:
            true,       // reload:
            false,      // onLoadEvent:
            "_self",        // target:
            0,              // triggeringEvent:
            0,          // form:
            0);         // formValues:
}

void FrameImpl::stopLoad()
{
    frameLoadDelegate()->onFrameLoadFinished(this, 1);
    closeURL();
}


void FrameImpl::emitFrameLoadStarted()
{
    frameLoadDelegate()->onFrameLoadStarted(this);
}

void FrameImpl::emitFrameLoadFinished(int status)
{
    frameLoadDelegate()->onFrameLoadFinished(this, status);
}

void FrameImpl::emitResourceLoadStarted(const OSB::ResourceStatus * status)
{
    resourceLoadDelegate()->onResourceLoadStarted(this, status);
}

void FrameImpl::emitResourceLoadHeaders(const OSB::ResourceStatus * status)
{
    resourceLoadDelegate()->onResourceLoadHeaders(this, status);
}

void FrameImpl::emitResourceLoadStatus(const OSB::ResourceStatus * status)
{
    resourceLoadDelegate()->onResourceLoadStatus(this, status);
}

void FrameImpl::emitResourceLoadFinished(const OSB::ResourceStatus * status)
{
    resourceLoadDelegate()->onResourceLoadFinished(this, status);
}

void FrameImpl::emitTitleChanged(const gchar* newtitle)
{
    assignToString(&_title,newtitle);
    frameLoadDelegate()->onTitleChanged(this, newtitle);
}

void FrameImpl::emitSetStatusText(const gchar* newtext)
{
	if (getUIDelegate())
		webkit_ui_delegate_set_status_text(getUIDelegate(), this->root()->api(), newtext);
}

void FrameImpl::emitClientRedirectReceived(const gchar* url)
{
    frameLoadDelegate()->onClientRedirectReceived(this, url);
}

void FrameImpl::emitClientRedirectCancelled(const gchar* url)
{
    frameLoadDelegate()->onClientRedirectCancelled(this, url);
}

void FrameImpl::emitServerRedirected(const gchar* url)
{
    frameLoadDelegate()->onServerRedirected(this, url);
}

void FrameImpl::emitCommitLoad()
{
    frameLoadDelegate()->onCommitLoad(this);
}

void FrameImpl::emitSetCookie(const gchar* value, const gchar* url)
{
	GURI* uri = gnet_uri_new(url);

	// this is to support old type  signals
	CookieParser p;
	Cookie** av = p.parse(value);
	Cookie* cookie;

	for (int i=0; (cookie = av[i]); ++i) {
		if (cookie->isValidForURL(uri)) {
			if (frameLoadDelegate()->onSetCookie(this, cookie, url)) {
				CookieStorage::shared()->setValidCookieForURL(cookie, uri);
				av[i] = 0;
			}
		}
    }

	// these should go directly like
	// WebCoreCookieAdapter::sharedAdapter()->setCookiesForURL(value, url, 0);
	gnet_uri_delete(uri);
}



class ElementInfoImpl : public OSB::ElementInfo
{
    WebCoreElementInfo *info;
public:
    ElementInfoImpl(WebCoreElementInfo* ainfo):info(ainfo) {}
    ~ElementInfoImpl() {}

    gboolean selected()
        { return info->isSelected;}
    const gchar* linkTitle()
        { return info->linkTitle; }
    const gchar*  linkLabel()
        { return info->linkLabel; }
    const gchar* linkURL()
        { return info->linkURL; }
    const gchar* linkTarget()
        { return info->linkTarget;}

    /** x = y = w = h= -1 if there's no image */
    void imageRect(int &x, int &y, int &w,int &h)
        {
            x=info->imageRect.x;
            y=info->imageRect.y;
            w=info->imageRect.width;
            h=info->imageRect.height;
        }

    const gchar* imageURL()
        { return info->imageURL;}

    const gchar* imageAltText()
        { return info->imageAltText;}

};

void FrameImpl::emitSelectionChanged()
{
    uiDelegate()->onSelectionChanged(this);
}

void FrameImpl::runJavaScriptAlertPanelWithMessage(const gchar* message)
{
    if (getUIDelegate())
		webkit_ui_delegate_run_javascript_alert_panel_with_message(getUIDelegate(), root()->api(), message, api());
}

bool FrameImpl::runJavaScriptConfirmPanelWithMessage(const gchar* message)
{
    if (getUIDelegate())
		return webkit_ui_delegate_run_javascript_confirm_panel_with_message(getUIDelegate(), root()->api(), message, api());
	return FALSE;
}

void FrameImpl::emitAuthenticate(const gchar* URL,
                                 const gchar* realm,
                                 gboolean* shouldTryAuth,
                                 gchar** username,
                                 gchar** password)
{
    assert(username);
    assert(password);
    bool ret = uiDelegate()->authPanel(this, URL, realm, username, password);
    *shouldTryAuth = ret;
}

bool FrameImpl::runJavaScriptTextInputPanelWithPrompt(const gchar* prompt,
                                                      const gchar* defaultText,
                                                      gchar** result)
{
    if (getUIDelegate())
		*result = webkit_ui_delegate_run_javascript_input_panel_with_prompt(getUIDelegate(), root()->api(), prompt, defaultText, api());
	return FALSE;
}

void FrameImpl::addMessageToConsole(const gchar* message, unsigned lineNumber, const gchar* sourceURL)
{
    //uiDelegate()->consoleMessage() ...

    g_printerr("javascript console message: '%s', line %d url %s",
               message,
               lineNumber,
               sourceURL);

}

GList* FrameImpl::peers()
{
    return root()->peers();
}

bool FrameImpl::areToolbarsVisible()
{
    return root()->features()->areToolbarsVisible();
}
void FrameImpl::setToolbarsVisible(bool visible)
{
    root()->features()->setToolbarsVisible(visible);
}

bool FrameImpl::isStatusBarVisible()
{
    return root()->features()->isStatusBarVisible();
}

void FrameImpl::setStatusBarVisible(bool visible)
{
    root()->features()->setStatusBarVisible(visible);
}

bool FrameImpl::areScrollbarsVisible()
{
    return root()->features()->areScrollbarsVisible();
}

void FrameImpl::setScrollbarsVisible(bool visible)
{
    root()->features()->setScrollbarsVisible(visible);
}

bool FrameImpl::windowIsResizable()
{
    return root()->features()->windowIsResizable();
}

void FrameImpl::setWindowIsResizable(bool resizable)
{
    root()->features()->setWindowIsResizable(resizable);
}

void FrameImpl::showWindow()
{
	if (getUIDelegate())
		webkit_ui_delegate_show_view(getUIDelegate(), root()->api());
}

void FrameImpl::closeWindowSoon()
{
	if (getUIDelegate())
		webkit_ui_delegate_close_view (getUIDelegate(), root()->api());
}

void FrameImpl::setWindowFrame(GdkRectangle* frame)
{
    GtkRequisition size;
    size.width = frame->width;
    size.height = frame->height;
    uiDelegate()->setWindowSize(this, &size, false);
}

void FrameImpl::windowFrame(GdkRectangle* outFrame)
{
    GtkAllocation allocation;
    uiDelegate()->getWindowSize(this, &allocation, false);
    fillGdkRectangle(outFrame, 0, 0, 0, 0);
//   fillGdkRectangle(outFrame, allocation.x, allocation.y,
//                    allocation.width, allocation.height);
}

void FrameImpl::setWindowContentRect(GdkRectangle* frame)
{
    GtkRequisition size;
    size.width = frame->width;
    size.height = frame->height;
    uiDelegate()->setWindowSize(this, &size, true);
}


void FrameImpl::windowContentRect(GdkRectangle* outRect)
{
    GtkAllocation allocation;
    uiDelegate()->getWindowSize(this, &allocation, true);
    fillGdkRectangle(outRect, 0, 0, 0, 0);
//   fillGdkRectangle(outRect, allocation.x, allocation.y,
//                    allocation.width, allocation.height);
}

void FrameImpl::setUserAgentString(const gchar* str)
{
    if (_userAgent)
        g_free (_userAgent);

    _userAgent = g_strdup(str);
}

const gchar* FrameImpl::userAgentForURL(const gchar* URL)
{
    if (_userAgent)
        return _userAgent;

    if (root()->mainFrame() != this)
        return root()->userAgentForURL(URL);

    return BridgeImpl::userAgentForURL(URL);
}
