/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * Copyright (c) 2004 Nokia. All rights reserved.
 * Copyright (c) 2007 Kimmo Kinnunen. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of Nokia nor the names of its contributors may be
 * used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <assert.h>
#include <string.h>
#include <glib.h>
#include <curl/curl.h>
#include "CurlFactory.h"
#include "CurlRequest.h"

using namespace Loader;

HTTPFactory& Loader::getHTTPFactory()
{
    static CurlFactory curlFactory;
    return curlFactory;
}

extern "C" {
static gboolean event_cb(gpointer user_data)
{
	assert(user_data);
	CurlFactory* self = static_cast<CurlFactory*>(user_data);
	return self->event();
}
}


static bool curlInitialized = false;

CurlFactory::CurlFactory()
{
    if (!curlInitialized) {
		curl_global_init(CURL_GLOBAL_ALL);
		curlInitialized = true;
    }

    curl_version_info_data *curl_info;
    curl_info = curl_version_info(CURLVERSION_NOW);
    curl_protocols = curl_info->protocols;

    FD_ZERO(&fdread);
    FD_ZERO(&fdwrite);
    FD_ZERO(&fdexcep);

    gsId = g_timeout_add_full(G_PRIORITY_DEFAULT_IDLE,
							  callbackTimeoutMS, // milliseconds
							  (GSourceFunc)& ::event_cb, // callback
							  this, // data
							  0); //notify

    curlMultiHandle = curl_multi_init();
}

CurlFactory::~CurlFactory()
{
    if (gsId) {
		GSource* s = g_main_context_find_source_by_id(0, gsId);
		if (s) g_source_destroy(s);
    }

    if (curlMultiHandle)
		curl_multi_cleanup(curlMultiHandle);

    if (curlInitialized) {
		curl_global_cleanup();
		curlInitialized = false;
    }
}

void CurlFactory::setProxy(const gchar* proto, const gchar* proxyURL)
{
    m_proxy = proxyURL;
}

bool CurlFactory::canProvide(const gchar * url) const
{
    if (!url) return false;

    int i = 0;
    while(curl_protocols[i]) {
		if (g_str_has_prefix(url, curl_protocols[i]))
			return true;
		++i;
    }

    return false;
}

Request* CurlFactory::createRequest(RequestListener* aListener,
                                    HTTPHeaderListener* headerListener,
                                    HTTPAuthorizer* authorizer,
                                    const gchar* aUrl,
                                    const gchar* aCookies,
                                    bool isSynchronized,
                                    const gchar* postContentType,
                                    GByteArray* postData)
{
    if (!curlMultiHandle) {
		return 0;
    }

    CurlRequest* req =
		new CurlRequest(this,  aListener, headerListener, authorizer, aUrl, aCookies, isSynchronized, postContentType, postData);

    return req;
}

void CurlFactory::update()
{

    FD_ZERO(&fdread);
    FD_ZERO(&fdwrite);
    FD_ZERO(&fdexcep);
    curl_multi_fdset(curlMultiHandle, &fdread, &fdwrite, &fdexcep, &maxfd);
}


/** timeout event callback
 *  @return FALSE if timeout is to be removed,
 *          TRUE if timeout is to be occur again
 */
gboolean CurlFactory::event()
{
    update();

    int nrunning;
    struct timeval timeout;
    int retval;
    timeout.tv_sec = 0;
    timeout.tv_usec = selectTimeoutMS * 1000;   // select waits microseconds

    retval = ::select(maxfd+1, &fdread, &fdwrite, &fdexcep, &timeout);
    switch(retval) {
    case -1:            // select error
#if DEBUG
		g_printerr("%s, select error(?)",__PRETTY_FUNCTION__);
#endif
		/* fallthrough*/
    case 0:         // select timeout
		/* fallthrough. this can be the first perform to be made */
    default:            // 1+ descriptors have data
		while(CURLM_CALL_MULTI_PERFORM == curl_multi_perform(curlMultiHandle, &nrunning));
		break;
    }

    // check the curl messages indicating completed transfers
    // and free their resources
    CurlRequest* req;
    CURLMsg* msg;
    CURL *handle;
    int nmsgs;

    while ((msg = curl_multi_info_read(curlMultiHandle, &nmsgs)))
    {
		handle = msg->easy_handle;

		if (msg->msg == CURLMSG_DONE) {
			// find the node which has same handle as completed transfer
			assert(handle);

			curl_easy_getinfo(handle, CURLINFO_PRIVATE, &req);
			assert(req); //fixme: assert->if ?

			// if found, delete it
			if (req) {
				switch (msg->data.result) {
				case CURLE_OK:
				{
                    
					// use this to authenticate
					int respCode = -1;
					curl_easy_getinfo(handle, CURLINFO_RESPONSE_CODE, &respCode);
                    if (req->shouldReexecute()) {
                        req->execute();
                    } else if(respCode == HTTP_STATUS_AUTH_REQUIRED &&
                        req->shouldRetry()) {
						// reset -- be careful not to call finish.
						// another thing also is to validate whether
						// data should be shown prior
						req->reset();
						// execute query again.
						req->execute();
					} else {
						req->finish();
					}
					break;
				}
				default:
				{
					req->error(msg->data.result);
					req->finish();
					break;
				}
				}

			} else {
#if DEBUG
				g_printerr("CurlRequest not found, eventhough curl handle finished");
#endif
				assert(0);
			}
		}
    }

    return TRUE;
}
